#!/usr/bin/python
import os, sys, getopt

import gobject
import random
import dbus
import dbus.service

import manifest
import servertracker
import settings
import tempfile
from store import Store

def get_max_version(versions):
    max = 0
    for range in versions:
        if range[1] > max:
            max = range[1]
    return max

def random_key(map):
    global random
    
    keys = map.keys()
    if len(keys) == 0:
        return None
    i = 0
    if len(keys) > 1:
        i = random.randint(0, len(keys)-1)
    return keys[i]

def set_union(map1, map2):
    union = map1.copy()
    for i in map2:
        union[i] = 1
    return union

class AutoUpdatinator(dbus.service.Object):
    def __init__(self, current_manifest, current_version, current_image_dir = None):

        # TODO: Should this really be system bus?
        self.bus = dbus.SessionBus()

        dbus.service.Object.__init__(self, dbus.service.BusName("org.freedesktop.Updatinator", bus=self.bus), "/updatinator")
        
        self.current_manifest = current_manifest
        self.current_version = current_version
        self.current_image_dir = current_image_dir

        self.download_timeout = 0

        self.downloaded_version = 0
        self.downloaded_required = {}

        self.downloading_version = 0
        self.downloading_required = {}
        self.downloading_remaining = {}
        self.downloading_manifest = None

        self.store = Store(cleanup = True, current_image_dir = current_image_dir, current_manifest = current_manifest)

        # Find the current fully downloaded version
        versions = self.store.get_availible_versions(settings.image_id, reverse=True)
        for version in versions:
            if version <= self.current_version:
                break
            req = self.store.get_required_blobs_for_update_to_version (None, settings.image_id, version)
            if self.store.are_all_blobs_availible(req):
                self.downloaded_version = version
                self.downloaded_required = req
                break;

        # Do we have a newer partial download?
        if len(versions) > 0 and versions[0] > self.current_version and versions[0] > self.downloaded_version:
            self.downloading_version = versions[0]
            self.downloading_required = self.store.get_required_blobs_for_update_to_version (None, settings.image_id, self.downloading_version)
            self.downloading_remaining = self.store.blobs_not_availible (self.downloading_required)
            self.downloading_manifest = self.store.get_manifest_for_version(settings.image_id, self.downloading_version)

        print "current downloaded version: %d, downloading version: %d" % (self.downloaded_version, self.downloading_version)

        self.tracker = servertracker.ServerTracker()
        self.tracker.connect("server-added", self.server_added)
        self.tracker.connect("server-removed", self.server_removed)

    # DBUS Updatinator interface

    @dbus.service.method(dbus_interface='org.fedoraproject.Updatinator',
                         in_signature='', out_signature='s', utf8_strings=True)
    def GetImage(self):
        return settings.image_id
    
    @dbus.service.method(dbus_interface='org.fedoraproject.Updatinator',
                         in_signature='', out_signature='i')
    def GetCurrentVersion(self):
        return self.current_version
    
    @dbus.service.method(dbus_interface='org.fedoraproject.Updatinator',
                         in_signature='', out_signature='i')
    def GetDownloadingVersion(self):
        return self.downloading_version

    @dbus.service.method(dbus_interface='org.fedoraproject.Updatinator',
                         in_signature='', out_signature='i')
    def GetDownloadedVersion(self):
        return self.downloaded_version

    @dbus.service.signal(dbus_interface='org.fedoraproject.Updatinator',
                         signature='si')
    def VersionAvailible(self, image_id, version):
        print "Got all data for image %s, version %d" % (image_id, version)

    @dbus.service.signal(dbus_interface='org.fedoraproject.Updatinator',
                         signature='si')
    def StartedDownload(self, image_id, version):
        print "Started downloading data for image %s, version %d" % (image_id, version)

    @dbus.service.signal(dbus_interface='org.fedoraproject.Updatinator',
                         signature='sisii')
    def DownloadProgress(self, image_id, version, sha1, downloaded, total):
        print "Downloaded blob %s of %s.%d (%d of %d)" % (sha1, image_id, version, downloaded, total)

    # End of DBUS Updatinator interface
    
    def remove_unneeded_blobs(self):
        to_remove = self.store.get_availible_blobs()
        for sha1 in self.downloading_required:
            if to_remove.has_key(sha1):
                del to_remove[sha1]
        for sha1 in self.downloaded_required:
            if to_remove.has_key(sha1):
                del to_remove[sha1]
        for sha1 in to_remove:
            self.store.remove_blob(sha1)

    def remove_unneeded_diffs(self):
        union = self.store.current_manifest.get_all_blobs()
        if self.downloaded_required:
            union = set_union (union, self.downloaded_required);
        if self.downloaded_required:
            union = set_union (union, self.downloading_required);

        self.store.remove_unneeded_diffs(union)

    def remove_unneeded_files(self):
        self.remove_unneeded_blobs()
        self.remove_unneeded_diffs()

    def download_timeout_cb(self):
        if self.downloading_version == 0:
            self.download_timeout = 0
            return False

        if len (self.downloading_remaining) == 0:
            self.downloaded_version = self.downloading_version
            self.downloaded_required = self.downloading_required

            self.downloading_version = 0
            self.downloading_required = {}
            self.downloading_remaining = {}

            self.remove_unneeded_files ()

            self.VersionAvailible(settings.image_id, self.downloaded_version)

            self.download_timeout = 0
            return False

        sha1 = random_key(self.downloading_remaining)

        server = self.tracker.find_random_server_for (settings.image_id, self.downloading_version)
        if server == None:
            print "No more servers, waiting"
            self.download_timeout = 0
            return False

        print "downloading %s from %s" % (sha1, server.base_uri)
        if self.store.fetch_blob (server, sha1, self.downloading_manifest):
            del self.downloading_remaining[sha1]

        total = len(self.downloading_required)
        self.DownloadProgress(settings.image_id, self.downloading_version, sha1,
                              total - len (self.downloading_remaining), total)
        
        return True

    def start_download(self):
        if self.download_timeout != 0:
            return

        # TODO: Pick timeout value?
        self.download_timeout = gobject.timeout_add(1000, self.download_timeout_cb)

    def server_added(self, tracker, server, image, versions):
        if image != settings.image_id:
            return
        if not server.can_download_from():
            return

        v = get_max_version(versions)

        # Do we already have this version installed?
        if v <= self.current_version:
            return

        # Do we already have this version downloaded?
        if self.downloaded_version != 0 and v <= self.downloaded_version:
            return
        
        # Are we downloading this version atm?
        if self.downloading_version != 0:
            if v < self.downloading_version:
                return
            elif v == self.downloading_version:
                # Maybe we're waiting for servers with this version, trigger a download
                self.start_download()
                return

        print "Found manifest for version %d, downloading" % (v, )

        if not self.store.fetch_manifest (server, settings.image_id, v):
            return

        self.StartedDownload (settings.image_id, v)

        self.downloading_version = v
        self.downloading_required = self.store.get_required_blobs_for_update_to_version (None, settings.image_id, v)
        self.downloading_remaining = self.store.blobs_not_availible (self.downloading_required)
        self.downloading_manifest = self.store.get_manifest_for_version(settings.image_id, self.downloading_version)

        self.remove_unneeded_files()

        self.start_download ()
    
    def server_removed (self, tracker, server):
        pass


def usage(retval = 0):
    print "%s <current manifest> <current version> [current image]\n" % sys.argv[0]
    sys.exit(retval)


############ Main 

try:
    opts, args = getopt.getopt(sys.argv[1:], "", ["help", None])
except getopt.GetoptError:
    usage(2)    
        
for o, a in opts:
    if o in ("-h", "--help"):
        usage()

if len(args) != 2 and len(args) != 3:
    usage(2)

current_manifest = args[0]
current_version = int(args[1])
current_image = None
if len(args) >= 3:
    current_image = args[2]

updatinator = AutoUpdatinator(current_manifest, current_version, current_image)

try:
    gobject.MainLoop().run()
except KeyboardInterrupt, k:
    pass
