import gobject
import avahi, dbus, dbus.glib
import random
import server
import os

def print_error(err):
    print "Error:", str(err)

def random_key(map):
    global random
    
    keys = map.keys()
    if len(keys) == 0:
        return None
    i = 0
    if len(keys) > 1:
        i = random.randint(0, len(keys)-1)
    return keys[i]

def split_versions (version_string):
    versions = []
    ranges = version_string.split(",")
    for range in ranges:
        range_split = range.split("-")
        try:
            if len (range_split) == 1:
                v = int(range_split[0], 10)
                versions.append( (v, v) )
            else:
                start = int (range_split[0], 10)
                end = int (range_split[1], 10)
                if end >= start:
                    versions.append( (start, end) )
        except ValueError, k:
            pass
    return versions

def has_version (versions, version):
    for range in versions:
        if version >= range[0] and version <= range[1]:
            return True
    return False

class ServerTracker(gobject.GObject):
    __gsignals__ = { 
        'server-added' : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
                          (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        'server-removed' : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
                            (gobject.TYPE_PYOBJECT, ))
        }

    def __init__(self):
        gobject.GObject.__init__(self)

        self.service_browsers = {}
        self.availible_servers = {}
        
        self.bus = dbus.SystemBus()
        self.avahi = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER), avahi.DBUS_INTERFACE_SERVER)
        
        self.browse_domain(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC, "local")
        
        db = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, self.avahi.DomainBrowserNew(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC, "", avahi.DOMAIN_BROWSER_BROWSE, dbus.UInt32(0))), avahi.DBUS_INTERFACE_DOMAIN_BROWSER)
        db.connect_to_signal('ItemNew', self.new_domain)

    def new_domain(self, interface, protocol, domain, flags):
        # We browse for .local anyway...
        if domain != "local":
            self.browse_domain(interface, protocol, domain)
            
    def browse_domain(self, interface, protocol, domain):

        # Are we already browsing this domain
        if self.service_browsers.has_key( (interface, protocol, domain) ):
            return

        b = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, self.avahi.ServiceBrowserNew(interface, protocol, "_updatinator._tcp", domain, dbus.UInt32(0))), avahi.DBUS_INTERFACE_SERVICE_BROWSER)
        b.connect_to_signal('ItemNew', self.new_service)
        b.connect_to_signal('ItemRemove', self.remove_service)

        self.service_browsers[(interface, protocol, domain)] = b

    def new_service(self, interface, protocol, name, stype, domain, flags):
        # Asynchronous resolving
        self.avahi.ResolveService(interface, protocol, name, stype, domain, avahi.PROTO_UNSPEC, dbus.UInt32(0), reply_handler=self.service_resolved, error_handler=print_error)
            
    def remove_service(self, interface, protocol, name, stype, domain, flags):
        k = (int (interface), str(name), int(protocol))
        srvr = self.availible_servers.pop (k, None)
        if srvr != None:
            self.server_removed (srvr, k)

    def service_resolved(self, interface, protocol, name, stype, domain, host, aprotocol, address, port, txt, flags):
        # Convert from dbus types:
        interface = int(interface)
        protocol = int(protocol)
        name = str(name)
        address = str(address)
        port = int(port)
        txt = avahi.txt_array_to_string_array(txt)
        
        image = None
        versions = []
        base = ""
        prio = 1
        try_diffs = False
        
        for key in txt:
            if key.startswith("image="):
                image = key[6:]
            elif key.startswith("versions="):
                versions = split_versions (key[9:])
            elif key.startswith("base="):
                base = key[5:]
            elif key.startswith("prio="):
                try:
                    prio = int(key[5:])
                except ValueError, k:
                    pass
            elif key == "diff":
                try_diffs = True

        if not image:
            return # We need an image!

        if protocol == 1: # ipv6
            if address[0:4] == "fe80":
                return # Can't create uris for link local ipv6 addresses (no way to specify iface)
            uri = "http://[%s]:%d/" % (address, port)
        else:
            uri = "http://%s:%d/" % (address, port)
        uri = os.path.join (uri, base)
        
        srvr = server.Server(uri)
        srvr.set_try_diffs (try_diffs)
        srvr.set_protocol (protocol)
        
        key = (interface, name, protocol)
        self.availible_servers[key] = (srvr, image, versions, prio)
        self.server_added(srvr, image, versions, key)

    def server_added (self, srvr, image, versions, key):
        self.emit("server-added", srvr, image, versions)

    def server_removed (self, srvr, key):
        self.emit("server-removed", srvr)

    def find_random_server_for(self, image_id, version):
        # TODO: Should look at prio, and distance on mesh
        availible = {}
        for ((iface, name, protocol), (srvr, image, versions, prio)) in self.availible_servers.iteritems():
            if image == image_id and has_version(versions, version) and srvr.can_download_from():
               if not availible.has_key(name) or protocol == 1:  # prefer ipv6 (== 1)
                   availible[name] = srvr

        k = random_key(availible)
        if k == None:
            return None
        return availible[k]
                   

gobject.type_register(ServerTracker)

if __name__ == "__main__":
    tracker = ServerTracker()

    def server_added (tracker, srvr, image, versions):
        print "server added %s" % (str(srvr))
        
    def server_removed (tracker, srvr):
        print "server removed %s" % (str(srvr))
        
    tracker.connect("server-added", server_added)
    tracker.connect("server-removed", server_removed)
    
    try:
        gobject.MainLoop().run()
    except KeyboardInterrupt, k:
        pass
        
