import settings
from manifest import Manifest
from release import Release
import os, sys
import tempfile
try:
    import hashlib
    def new_sha1():
        return hashlib.sha1()
except ImportError:
    import sha
    def new_sha1():
        return sha.new()

def download_file (in_file, out_file):
    while True:
        data = in_file.read(8192)
        if data == "":
            break
        out_file.write(data)
    out_file.flush()

def sha1_file(path):
    f = None
    try:
        f = open(path, "rb")
        h = new_sha1()
        while True:
            data = f.read(4096)
            if data == "":
                break
            h.update(data)
        f.close()
        return h.hexdigest()
    except:
        if f:
            f.close()
        return None

def find_in_path(file):
    path = os.environ.get("PATH")
    if path:
        for dir in path.split(":"):
            p = os.path.join(dir, file)
            if os.path.isfile(p):
                # TODO: Should test executable bit
                return True
    return False

BLOB_NORMAL, BLOB_COMPRESSED = 1, 2

class Store:
    def __init__(self, cleanup = False, current_manifest = None, current_image_dir = None):
        self.cleanup = cleanup
        self.availible_blobs = {}
        self.availible_versions = {}
        self.current_manifest = None
        self.current_manifest_sha1 = None
        if current_manifest:
            self.current_manifest = Manifest(current_manifest)
            self.current_manifest_sha1 = sha1_file(current_manifest)
            
        self.current_image_dir = current_image_dir

        self.use_diffs = current_manifest and find_in_path("bspatch")

        # Make sure we have these dirs
        if not os.path.isdir(settings.blob_dir):
            os.mkdir (settings.blob_dir)
        if not os.path.isdir(settings.release_dir):
            os.mkdir (settings.release_dir)
            
        if settings.save_diffs and not os.path.isdir(settings.diff_dir):
            os.mkdir (settings.diff_dir)

        # Find availible blobs in blob_dir
        for file in os.listdir(settings.blob_dir):
            if file[-1:] == "~":
                pass
            if file.startswith ("tmp-"):
                if cleanup:
                    os.remove(os.path.join(settings.blob_dir, file))
            elif len (file) == 40: # sha1 files are 40 chars
                sha1 = file
                v = 0
                if self.availible_blobs.has_key (sha1):
                    v = self.availible_blobs[sha1]
                self.availible_blobs[sha1] = v | BLOB_NORMAL
            elif len (file) == 43 and file[-3:] == ".gz": # gzip:ed sha1 files are 43 chars
                sha1 = file[:-3]
                v = 0
                if self.availible_blobs.has_key (sha1):
                    v = self.availible_blobs[sha1]
                self.availible_blobs[sha1] = v | BLOB_COMPRESSED

        # Find availible versions in release_dir
        for file in os.listdir(settings.release_dir):
            if file.startswith ("tmp-"):
                if cleanup:
                    os.remove(os.path.join(settings.release_dir, file))
            else:
                dot = file.rfind (".")
                if  dot!= -1:
                    try:
                        image = file[:dot]
                        version = int(file[dot+1:])
                        self.add_availible_version(image, version)
                    except ValueError, k:
                        pass

    def add_availible_version(self, image_id, version):
        if not self.availible_versions.has_key(image_id):
            self.availible_versions[image_id] = {}
        self.availible_versions[image_id][version] = 1

    def get_availible_images(self):
        return self.availible_versions.keys()

    def get_availible_versions(self, image_id, reverse=False):
        if not self.availible_versions.has_key(image_id):
            return []
        versions = self.availible_versions[image_id].keys()
        versions.sort(reverse=reverse)
        return versions

    def get_release_for_version(self, image_id, version):
        try:
            path = os.path.join (settings.release_dir, "%s.%d" % (image_id, version))
            return Release(path)
        except:
            return None

    def load_manifest (self, sha1):
        if self.availible_blobs.has_key(sha1):
            m = Manifest()
            m.read_file (self.read_blob(sha1))
            return m
        return None

    def load_manifest_for_release (self, release):
        if release.get_manifest_ref():
            return self.load_manifest (release.get_manifest_ref())
        return None
    
    def get_required_blobs_for_update_to_manifest (self, manifest):
        required = {}
        if self.current_manifest:
            required = self.current_manifest.required_for_update (manifest)
        else:
            required = manifest.get_all_blobs()
        return required

    def read_blob (self, sha1):
        p = os.path.join (settings.blob_dir, sha1)
        return open(p, "rb")

    def are_all_blobs_availible (self, blobs):
        for i in blobs:
            if not self.availible_blobs.has_key(i):
                return False
        return True

    def release_is_availible(self, image_id, version):
        if self.availible_versions.has_key(image_id):
            return self.availible_versions[image_id].has_key(version)
        return False

    def blob_is_availible(self, blob):
        return self.availible_blobs.has_key (blob)

    def get_availible_blobs(self):
        return self.availible_blobs.copy()

    def blobs_not_availible (self, required_blobs):
        remaining = {}
        for i in required_blobs:
            if not self.availible_blobs.has_key(i):
                remaining[i] = 1
        return remaining

    def remove_blob(self, sha1):
        if self.availible_blobs.has_key (sha1):
            sys.stderr.write ("removing stale blob %s\n" % (sha1, ))
            path = os.path.join (settings.blob_dir, sha1)
            os.remove(path)
            path = os.path.join (settings.blob_dir, sha1 + ".gz")
            os.remove(path)
            del self.availible_blobs[sha1]

    def remove_unneeded_diffs(self, required_blobs):
        for file in os.listdir(settings.diff_dir):
            if file.startswith ("tmp-"):
                os.remove(os.path.join(settings.diff_dir, file))
                
            elif len (file) == 86 and file[40] == "." and file[-5:] == ".diff":
                # sha1 files are 40+1+40+5 chars, ignore others
                from_sha1 = file[:40]
                to_sha1 = file[41:81]
                if not required_blobs.has_key(to_sha1):
                    os.remove(os.path.join(settings.diff_dir, file))

    def save_blob(self, sha1, in_file):
        (fd, tmp_path) = tempfile.mkstemp(dir=settings.blob_dir, prefix="tmp-")
        out_file = os.fdopen(fd, "wb")
        res = True

        try:
            h = new_sha1()
            while True:
                data = in_file.read(4096)
                if data == "":
                    break
                h.update(data)
                out_file.write(data)
        except:
            sys.stderr.write ("save_blob exception: %s\n" % (str(sys.exc_info()[1]), ))
            res = False
        out_file.close()

        res_sha1 = h.hexdigest()
        if not res or res_sha1 != sha1:
            os.unlink (tmp_path)
            return False

        os.rename (tmp_path, os.path.join (settings.blob_dir, sha1))
        self.availible_blobs[sha1] = 1
        return True

    def save_diff(self, from_sha1, to_sha1, in_file):
        (fd, tmp_path) = tempfile.mkstemp(dir=settings.diff_dir, prefix="tmp-")
        out_file = os.fdopen(fd, "w")
        res = True

        try:
            while True:
                data = in_file.read(4096)
                if data == "":
                    break
                out_file.write(data)
        except:
            sys.stderr.write("save_diff exception: %s\n" % (str(sys.exc_info()[1]), ))
            res = False
            os.unlink (tmp_path)
        out_file.close()
        
        if res:
            # We rely on always getting ok diffs, we verify that they apply before saving them
            os.rename (tmp_path, os.path.join (settings.diff_dir, "%s-%s.diff" % (from_sha1, to_sha1)))
        return res

    def save_release(self, image_id, version, in_file):
        filename =  "%s.%d" % (image_id, version)
        res = True
        (fd, tmp_path) = tempfile.mkstemp(dir=settings.release_dir, prefix="tmp-")

        try:
            out_file = os.fdopen(fd, "w")
            while True:
                data = in_file.read(4096)
                if data == "":
                    break
                out_file.write(data)
        except:
            sys.stderr.write("save_manifest exception: %s\n" % (str(sys.exc_info()[1]), ))
            res = False
            os.unlink (tmp_path)
        out_file.close()

        if res:
            # TODO: Check gpg signed file, if failed, return False ans set failed on server
            os.rename (tmp_path, os.path.join (settings.release_dir, filename))
            self.add_availible_version(image_id, version)

        return res

    def fetch_release (self, server, image_id, version):
        res = False
        in_file = None
        try:
            in_file = server.open_release(image_id, version)
            res = self.save_release (image_id, version, in_file)
        except:
            sys.stderr.write ("Unable to load release %s.%d from server %s\n"%(image_id, version, server.base_uri))
            sys.stderr.write ("Exception: %s" % (str(sys.exc_info()[1]), ))
        if in_file:
            in_file.close()

        if not res:
            server.set_failed() # One file failed, we don't trust this server anymore

        return res
    
    def fetch_blob_using_diff(self, server, from_path, from_sha1, to_sha1):
        diff_file = server.open_diff(from_sha1, to_sha1)
        if diff_file == None:
            return False

        # Download diff
        print "Downloading diff from %s to %s" % (from_sha1, to_sha1)
        diff_temp = tempfile.NamedTemporaryFile(mode="rw+b")
        download_file(diff_file, diff_temp)

        # Apply diff
        blob_temp = tempfile.NamedTemporaryFile(mode="r+b")
        args = ["bspatch", from_path, blob_temp.name, diff_temp.name]
        res = os.spawnvp(os.P_WAIT, "bspatch", args) == 0
        
        if res:
            res = self.save_blob(to_sha1, blob_temp)
            if res and settings.save_diffs:
                diff_temp.seek (0)
                self.save_diff(from_sha1, to_sha1, diff_temp)

        # The failure here could be a problem with the local file being modified.
        # Don't set server.failed, instead fall back to trying the blob
            
        diff_file.close()
        blob_temp.close()

        return res
        
    def fetch_blob (self, server, sha1, downloading_manifest = None, is_manifest = False):
        if self.use_diffs and server.try_diffs:
            if downloading_manifest != None:
                name = downloading_manifest.find_name_for_blob(sha1)
                if name:
                    original_blob = self.current_manifest.get_blob_for_name(name)
                    original_path = None
                    if self.availible_blobs.has_key(original_blob):
                        original_path = os.path.join (settings.blob_dir, original_blob)
                    elif self.current_image_dir:
                        original_path = os.path.join(self.current_image_dir, name)

                    if original_path and os.path.isfile(original_path):
                        res = self.fetch_blob_using_diff(server, original_path, original_blob, sha1)
                        if res:
                            return res # Succeeded with diff

            if is_manifest and self.current_manifest_sha1:
                original_path = os.path.join (settings.blob_dir, self.current_manifest_sha1)
                if os.path.isfile(original_path):
                    res = self.fetch_blob_using_diff(server, original_path, self.current_manifest_sha1, sha1)
                    if res:
                        return res # Succeeded with diff
                    
            # Fall back to direct blob getting

        res = False
        in_file = None
        try:
            in_file = server.open_blob(sha1)
            res = self.save_blob(sha1, in_file)
            in_file.close()
        except:
            sys.stderr.write ("fetch_blob exception: %s\n" % (str(sys.exc_info()[1]), ))
            sys.stderr.write ("Unable to load blob %s from %s\n"%(sha1, server.base_uri))
            
        if in_file:
            in_file.close()

        if not res:
            server.set_failed() # One file failed, we don't trust this server anymore

        return res
        
