#!/usr/bin/python

from manifest import Manifest
from release import Release
import os, sys
import gzip
import tempfile

def open_blob (sha1):
    global blob_dir
    path = os.path.join(blob_dir, sha1)
    if os.path.isfile(path):
        return open(path, "rb")
    path = path + ".gz"
    if os.path.isfile(path):
        return gzip.GzipFile(path, "rb")
    return None

def temp_unzip(zip_path):
    temp_file = tempfile.NamedTemporaryFile()
    in_file = gzip.GzipFile(zip_path, "rb")
    while True:
        data = in_file.read(8192)
        if data == "":
            break
        temp_file.write(data)
    temp_file.flush()
    in_file.close()
    return temp_file

def generate_diff(from_sha1, to_sha1):
    global diff_dir, blob_dir
    file = "%s-%s.diff"%(from_sha1, to_sha1)
    diff_path = os.path.join(diff_dir, file)
    if not os.path.isfile(diff_path):
        t1 = None
        t2 = None
        from_blob = os.path.join(blob_dir, from_sha1)
        if not os.path.isfile(from_blob):
            zip_path = from_blob + ".gz"
            if not os.path.isfile(zip_path):
                print "Unable to find blob %s" % (from_sha1, )
                return
            t1 = temp_unzip(zip_path)
            from_blob = t1.name


        to_blob = os.path.join(blob_dir, to_sha1)
        if not os.path.isfile(to_blob):
            zip_path = to_blob + ".gz"
            if not os.path.isfile(zip_path):
                print "Unable to find blob %s" % (to_sha1, )
                return
            t2 = temp_unzip(zip_path)
            to_blob = t2.name

        res = os.spawnvp(os.P_WAIT, "bsdiff", ["bsdiff", from_blob, to_blob, diff_path])

        if t1:
            t1.close()
        if t2:
            t2.close()


def generate_diffs(from_release, to_release):
    b1 = open_blob(from_release.get_manifest_ref())
    if not b1:
        print "Can't find manifest for release %s" % (from_release, )
        return
    m1 = Manifest(file = b1)
    b1.close()
    b2 = open_blob(to_release.get_manifest_ref())
    if not b2:
        print "Can't find manifest for release %s" % (to_release, )
    m2 = Manifest(file = b2)
    b2.close()

    generate_diff (from_release.get_manifest_ref(), to_release.get_manifest_ref())
    changes = m1.blobs_changed_in_update (m2)
    for change in changes:
        generate_diff (change[0], change[1])

def find_in_path(file):
    path = os.environ.get("PATH")
    if path:
        for dir in path.split(":"):
            p = os.path.join(dir, file)
            if os.path.isfile(p):
                # TODO: Should test executable bit
                return True
    return False

def setup_paths(base_dir, need_diff_dir = False):
    global release_dir, blob_dir, diff_dir
    release_dir = os.path.join(base_dir, "releases")
    blob_dir = os.path.join(base_dir, "blobs")
    diff_dir = os.path.join(base_dir, "diffs")

    if need_diff_dir and not os.path.isdir(diff_dir):
        os.mkdir (diff_dir)

def ensure_bsdiff():
    if not find_in_path("bsdiff"):
        print "Can't find required program 'bsdiff' is PATH"
        sys.exit(1)

def enumerate_images():
    global release_dir, blob_dir, diff_dir
    
    images = {}
    for file in os.listdir(release_dir):
        if not file.startswith ("tmp-") and file[-1:] != "~":
            i = file.find(".")
            if i > 0:
                image = file[:i]
                if not images.has_key(image):
                    images[image] = []
                try:
                    version = int(file[len(image)+1:])
                    path = os.path.join(release_dir, file)
                    r = Release(path)
                    images[image].append( (version, r))
                except:
                    print "Error loading release for file %s: %s" % (file, str(sys.exc_info()))
    return images

def generate_consecutive_diffs():
    global release_dir, blob_dir, diff_dir

    images = enumerate_images()

    for image in images:
        versions = images[image]
        versions.sort()
        i = 0
        # Don't include the last version as that has no diffs (yet)
        while i < len(versions) - 1:
            print "changes from version %d to %d in %s" % (versions[i][0], versions[i+1][0], image)
            generate_diffs (versions[i][1], versions[i+1][1])
            i = i + 1

def get_release(image):
    global release_dir
    p = image
    if not os.path.isfile(p):
        p = os.path.join(release_dir, p)
    if not os.path.isfile(p):
        print "Can't find image %s" % (image, )
    r = Release(p)
    return r

def compress_blob(blob, remove=False):
    global blob_dir
    p = os.path.join(blob_dir, blob)

    if not os.path.isfile(p):
        print "warning, no blob for %s" % (blob, )
        return

    in_file = file(p, "r")
    
    p2 = os.path.join(blob_dir, blob + ".gz")
    if os.path.isfile(p2):
        return # Already exists
    
    (fd, tmp_path) = tempfile.mkstemp(dir=blob_dir, prefix="tmp-")
    tmp_file = os.fdopen(fd, "w")

    out_file = gzip.GzipFile(blob, "w", 9, tmp_file)

    res = True
    try:
        while True:
            data = in_file.read(4096)
            if data == "":
                break
            out_file.write(data)
    except:
        print "compress_blob exception: %s" % (str(sys.exc_info()[1]), )
        res = False
    out_file.close()

    if not res:
        os.unlink (tmp_path)
        return False

    os.rename (tmp_path, p2)
    os.chmod(p2, 0644)
    
    if remove:
        os.unlink(p)
        
    return True

def compress_all(remove = False):
    global blob_dir
    for file in os.listdir(blob_dir):
        if len(file) == 40 and not file.startswith ("tmp-"):
            compress_blob(file, remove=remove)
    
def compress_image(r, remove = False):
    compress_blob(r.get_manifest_ref(), remove=remove)
    b = open_blob(r.get_manifest_ref())
    if not b:
        print "Can't find manifest for release %s" % (r, )
        return
    
    m = Manifest(file = b)
    b.close()
    for blob in m.get_all_blobs():
        compress_blob(blob, remove=remove)
    
def usage(retval = 0):
    print "Uses:"
    print "%s gendiffs <base directory>\n" % sys.argv[0]
    print "%s gendiff <base directory> <image1>.<version> <image2>.<version>\n" % sys.argv[0]
    print "%s compress-image <base directory> <image1>.<version>\n" % sys.argv[0]
    print "%s compress-all <base directory>\n" % sys.argv[0]
    sys.exit(retval)

args = sys.argv[1:]

if len(args) == 0:
    usage()

if args[0] == "gendiffs" and len(args) == 2:
    base_path = args[1]
    ensure_bsdiff()
    setup_paths(base_path, need_diff_dir = True)
    generate_consecutive_diffs()
    
elif args[0] == "gendiff" and len(args) == 4:
    base_path = args[1]
    release1 = args[2]
    release2 = args[3]

    ensure_bsdiff()
    setup_paths(base_path, need_diff_dir = True)

    r1 = get_release (release1)
    r2 = get_release (release2)
    generate_diffs (r1, r2)
    
elif args[0] == "compress-all" and len(args) == 2:
    base_path = args[1]

    setup_paths(base_path)
    
    compress_all(remove=True)
    
elif args[0] == "compress-image" and len(args) == 3:
    base_path = args[1]
    release = args[2]

    setup_paths(base_path)
        
    r = get_release (release)
    compress_image(r, remove=True)
    
else:
    usage()
