#!/usr/bin/python
import os, sys, getopt
import tempfile

import manifest
from server import Server
import settings
from store import Store

def usage(retval = 0):
    sys.stderr.write("%s  [-m <current manifest>] [-i <current image path>] [-u <base uri>] [--availible | --get-release <image> <version> | --cat-manifest <image> <version> | --download <image> <version>] \n" % sys.argv[0])
    sys.exit(retval)

def print_store(store):
    images = store.get_availible_images()
    for image in images:
        print "image '%s'" % (image)
        versions = store.get_availible_versions(image)
        for v in versions:
            rel = store.get_release_for_version(image, v)
            if not rel:
                print " %d - can't read release file" % (v, )
            else:
                m = store.load_manifest_for_release (rel)
                if not m:
                    print " %d - lacks manifest" % (v, )
                else:
                    required = store.get_required_blobs_for_update_to_manifest(m)
                    not_availible = store.blobs_not_availible(required)
                    if len(not_availible) == 0:
                        print " %d - complete" % (v, )
                    else:
                        print " %d - %d blobs missing" % (v, len(not_availible))

def fetch_release(image, version):
    if not store.release_is_availible(image, version):
        res = store.fetch_release (server, image, version)
        if not res:
            sys.stderr.write("Error downloading manifest\n")
            sys.exit(2)

    rel = store.get_release_for_version(image, version)
    if not rel:
        sys.stderr.write("Error in release file\n")
        sys.exit(2)
        
    return rel

def fetch_manifest(rel):
    manifest_sha1 = rel.get_manifest_ref()
    if manifest_sha1 and not store.blob_is_availible(manifest_sha1):
        if not store.fetch_blob (server, manifest_sha1, is_manifest = True):
            sys.stderr.write("Failed to fetch manifest %s\n" % (manifest_sha1, ))
            sys.exit(2)

def get_release(store, args):
    if len(args) != 2:
        usage(2)
        
    image = args[0]
    version = int(args[1])

    rel = fetch_release(image, version)
    fetch_manifest(rel)

    print "Release data for '%s' version '%d' availible" % (image, version)

def cat_manifest(store, args):
    if len(args) != 2:
        usage(2)
        
    image = args[0]
    version = int(args[1])

    rel = fetch_release(image, version)
    fetch_manifest(rel)
    
    manifest_sha1 = rel.get_manifest_ref()
    path = os.path.join (settings.blob_dir, manifest_sha1)

    f = open(path, "rb")
    while True:
        s = f.read(4096)
        if s=="":
            break
        sys.stdout.write(s)
        
    f.close()
    return


def download(store, args):
    if len(args) != 2:
        usage(2)
        
    image = args[0]
    version = int(args[1])

    rel = fetch_release (image, version)
    fetch_manifest(rel)

    manifest_sha1 = rel.get_manifest_ref()

    m = store.load_manifest_for_release(rel)
    required = store.get_required_blobs_for_update_to_manifest(m)
    not_availible = store.blobs_not_availible(required)

    l = len(not_availible)
    if l > 0:
        print "Downloading %d new blobs needed for update"%(l,)
    i = 0
    for blob in not_availible:
        print "Fetching blob %s (%d/%d)" % (blob, i+1, l)
        if not store.fetch_blob (server, blob, downloading_manifest=m):
            print "Failed to fetch blob %s" % (blob, )
            sys.exit(2)
        i = i + 1;

############ Main 

try:
    opts, args = getopt.getopt(sys.argv[1:], "hm:i:u:", ["help", "manifest", "imagedir", "uri", "availible", "get-release", "cat-manifest", "download"])
except getopt.GetoptError:
    usage(2)    

current_image_dir = None
current_manifest = None
server = None

opt_availible = False
opt_getrelease = False
opt_download = False
opt_cat_manifest = False

for o, a in opts:
    if o in ("-h", "--help"):
        usage()

    if o in ("-i", "--current-imagedir"):
        current_image_dir = a
    
    if o in ("-m", "--current-manifest"):
        current_manifest = a

    if o in ("-u", "--uri"):
        server = Server(a)

    if o == "--availible":
        opt_availible = True

    if o == "--get-release":
        opt_getrelease = True

    if o == "--download":
        opt_download = True

    if o == "--cat-manifest":
        opt_cat_manifest = True

if not server:
    server = Server(settings.download_uri)
    
store = Store(cleanup = False, current_image_dir = current_image_dir, current_manifest = current_manifest)

if opt_availible:
    print_store(store)
elif opt_getrelease:
    get_release(store, args)
elif opt_cat_manifest:
    cat_manifest(store, args)
elif opt_download:
    download(store, args)
else:
    usage()
